/*
 *
 *	@(#)config.h	1.2	94/03/06	(c) PACT
 *
 *	PACT Parallel C / transputer
 *
 *	Copyright (c) 1991-1994 PACT
 *	All Rights Reserved
 *
 */

#ifndef	__config_included
#define	__config_included


/* Channel type depends on compiler mode:								*/
#ifndef	CHAN
#	if	__PACT__ > 1
#		define	CHAN	void |
#	else
#		define	CHAN	int
#	endif
#endif


/* Nil pointer used to end lists:										*/
#ifndef	NIL
#	define	NIL		((void *) (-1))
#endif


/* Processor description, setup by configurer and patched by init code:	*/
struct _Processor {
	int					id;					/* unique number */
	int					mode;				/* mode */
	int					memsize;			/* total memory size */
	void				*code;				/* start of code memory */
	void				*data;				/* start of data memory */
	CHAN				*extch;				/* start of external channels */
	int					linkin;				/* input link */
	int					linkout;			/* output link */
	void				*freemem;			/* start of free memory */
	struct _Process		*uprocs;			/* list of user processes */
	struct _Process		*iprocs;			/* list of initialization processes */
	struct _Virtual		*virdata;			/* virtual data address */
	int					nofchans;			/* number of channels */
/*	void				|chans[];	*/		/* channels */
};

#define	PROCESSOR_SIZE	13					/* 13 words + whatever in chans[] */


/* Processor modes:														*/
#define	MODE_HOE		0x01				/* set halt-on-error flag */
#define	MODE_CFG		0x02				/* created by configurer */


/* Process description, setup by configurer:							*/
struct _Process {
	int					id;					/* unique number */
	int					mode;				/* mode */
	int					codesize;			/* size of code segment */
	void				*code;				/* start of code segment */
	void				*entry;				/* code entry point */
	int					stacksize;			/* size of stack */
	int					stackparmsize;		/* ... */
	void				*stack;				/* address of (top of) stack */
	int					vectorsize;			/* size of vector space */
	void				*vector;			/* and location */
	int					datasize;			/* size of data segment */
	void				*data;				/* and location */
	int					heapsize;			/* size of heap */
	void				*heap;				/* and start */
	struct _Process		*next;				/* next process in list */
	struct _Processor	*processor;			/* processor structure */
	int					nofparms;			/* # process parameters */
	int					nofvalues;			/* # process values */
/*	struct _Parm		parms[];		*/	/* parm info stored here */
/*	int					values[];		*/	/* value here */
};

#define	PROCESS_SIZE	18					/* 18 words + whatever in parms[] and values[] */


/* Process modes:														*/
#define	MODE_HIPRI		0x01				/* high priority? */
#define	MODE_SEQ		0x02				/* sequential? */
#define	MODE_STACK		0x04				/* separate stack? */
#define	MODE_DATA		0x1000				/* global data loaded (PACT style)? */
#define	MODE_SMS		0x2000				/* can stack grow down to MemStart? */


/* Parameter info in Process structure:									*/
struct _Parm {
	int		simple;							/* simple type */
	int		complete;						/* complete type */
	int		value;							/* value */
};

/* Simple types:														*/
#define	PARM_DATA		1
#define	PARM_DATAP		2
#define	PARM_INTERNAL	3
#define	PARM_EXTERNAL	4
#define	PARM_CHANNELP	5

/* Complete types:														*/
#define	PARM_BOOL		24		/* simple: DATA							*/
#define	PARM_BYTE		1		/*         DATA							*/
#define	PARM_INT		2		/*         DATA							*/
#define	PARM_INT16		3		/*         DATA							*/
#define	PARM_INT32		4		/*         DATA, DATA*					*/
#define	PARM_INT64		5		/*         DATA*						*/
#define	PARM_REAL32		6		/*         DATA, DATA*					*/
#define	PARM_REAL64		7		/*         DATA*						*/
#define	PARM_BOOLP		25		/*         DATA*						*/
#define	PARM_BYTEP		8		/*         DATA*						*/
#define	PARM_INTP		9		/*         DATA*						*/
#define	PARM_INT16P		10		/*         DATA*						*/
#define	PARM_INT32P		11		/*         DATA*						*/
#define	PARM_INT64P		12		/*         DATA*						*/
#define	PARM_REAL32P	13		/*         DATA*						*/
#define	PARM_REAL64P	14		/*         DATA*						*/
#define	PARM_INPUT		15		/*         INTERNAL, EXTERNAL, CHANNEL*	*/
#define	PARM_OUTPUT		16		/*         INTERNAL, EXTERNAL, CHANNEL*	*/
#define	PARM_RETURN		17		/*         DATA*						*/
#define	PARM_PROCESS	18		/*         DATA*						*/
#define	PARM_VECTOR		19		/*         DATA*						*/
#define	PARM_CODE		20		/*         DATA							*/
#define	PARM_SIZE		21		/*         DATA							*/
#define	PARM_FREE		22		/*         DATA*						*/
#define	PARM_STACK		23		/*         DATA*						*/


/* Processor capabilities set:											*/
#define	INSTR_CORE				0x0001
#define	INSTR_FMUL				0x0002
#define	INSTR_FP_SUPPORT		0x0004
#define	INSTR_DUP				0x0008
#define	INSTR_WSUBDB			0x0010
#define	INSTR_MOVE2D			0x0020
#define	INSTR_CRC				0x0040
#define	INSTR_BITOPS			0x0080
#define	INSTR_FPU_CORE			0x0100
#define	INSTR_FPTESTERR			0x0200
#define	INSTR_LDDEVID			0x0400
#define	INSTR_DEBUG_SUPPORT		0x0800
#define	INSTR_TIMER_DISABLE		0x1000
#define	INSTR_LDMEMSTARTVAL		0x2000
#define	INSTR_POP				0x4000


#endif	/* __config_included */

