#include <stdio.h>
#include <assert.h>
#include <conc.h>
#include <sys/crt.h>
#include <node.h>
#include "OS.h"

#define SECONDARY_WORKSPACE_SIZE	256
#define MAX_PARAMETER_LEN		1024
#define BOOT_BUFFER_SIZE	(1024*10)
#pragma PACT stackcheck off

#define NOT_PROCESS			0x80000000
#define MAX_TASKS			2
#define START_TASK			0
#define EXIT_TASK			1



/*	Network Loader parameters
--  Parameter 0 --> Pointer to return instruction
--  Parameter 1 --> Pointer to start of (network loader) block of code
--  Parameter 2 --> (Network loader) Code size
--  Parameter 3 --> Data load start	(mint)
--  Parameter 4 --> Code load start	(mint)
--  Parameter 5 --> Parent channel in
--  Parameter 6 --> Parent channel out
--  Parameter 7 --> Children channels in out
--  Parameter 8 --> Children channels in out size
--  Parameter 9 --> Pass on buffer area
--  Parameter 10--> Labelling block area
--
*/


/*	Fill the Process Info structure
*/
#define FILL_PROCESS_INFO( ProcessInfo, DataLo, DataSize, TextLo, TextSize,\
StackLo, StackSize, HeapLo, HeapSize, Entry, TaskId, NumParams ) {	\
	ProcessInfo->id = TaskId;					\
	ProcessInfo->mode = 3000;					\
	ProcessInfo->codesize = TextSize;				\
	ProcessInfo->code = (void *)TextLo;				\
	ProcessInfo->entry = (void *)Entry;				\
	ProcessInfo->stacksize = StackSize;				\
	ProcessInfo->stack = (void *)StackLo;				\
	ProcessInfo->vectorsize = 0;					\
	ProcessInfo->vector = NULL;					\
	ProcessInfo->datasize = DataSize;				\
	ProcessInfo->data = (void *)DataLo;				\
	ProcessInfo->heapsize = HeapSize;				\
	ProcessInfo->heap = (void *)HeapLo;				\
	ProcessInfo->next = (struct _Process *)NULL;			\
	ProcessInfo->nofparms = NumParams;				\
	ProcessInfo->nofvalues = 0;					\
}

extern void CallEntryPoint( int (*fn)(), unsigned long *mdl, unsigned long *ParameterBlock );
extern void _ChanIn( void * address, Channel * ParentInChan, int size );
extern int _VChanIn( void * address, Channel * ParentInChan, int size );
extern void _ChanOut( void * address, Channel * ParentOutChan, int size );


/*	COFF loader
*/
CoffLoader()
{
	void *stackFragmentInfo[MAX_TASKS][3];
	int *OldStack;
	struct coff_params {
		unsigned long TextStart;
		unsigned long TextSize;
		unsigned long DataStart;
		unsigned long DataSize;
		unsigned long BssStart;
		unsigned long BssSize;
		unsigned long NumSections;
		int (*entry)();
		int HeapSize;
		int StackSize;
		int DebugFlag;
	}CoffParameters;
	struct section_info {
		unsigned long SectionStart;
		unsigned long SectionSize;
	}SectionInfo;
	int i, len;
	unsigned long ParameterBlock[8];
	void *MDL[16];
	struct _Process ProcessInfo;
	int *NewStack;
	int *Parameters ;
	Channel *ParentInChan, *ParentOutChan;
	unsigned long addr;
	int *ptr;
	unsigned long TaskId;
	struct loader_info {
		int NodeId;
		int MemStart;
		int TotalMem;
		int NumVirtualChannels;
		long MinInternalChannel;
		int NumInternalChannels;
	}LoaderInfo;
	int *Params;
	struct fragments {
		long StartAddress;
		int Len;
		int Used;
		int TaskId;
	}Fragment[MAX_TASKS];
	struct task_info {
		int TaskId;
		int Action;
	}TaskInfo;
	int MemRequired;
	char *DebugCommand;

	ParentInChan =  ((Channel * (*) (const char *)) __asm)( "ldl .L1+5" );
	ParentOutChan =  ((Channel * (*) (const char *)) __asm)( "ldl .L1+6" );
	_VChanIn( &LoaderInfo, ParentInChan, sizeof(LoaderInfo) );
	OldStack = ((void * (*) (const char *)) __asm)( "ldlp 0" );
#if 0
	ptr = OldStack + SECONDARY_WORKSPACE_SIZE;
	_VChanIn( ptr, ParentInChan, LoaderInfo.NumVirtualChannels*8 );
	ConfigureUserVcp( LoaderInfo.NumVirtualChannels, ptr, ParentInChan );
#endif
	for ( ptr = (int *)(LoaderInfo.MinInternalChannel), i = 0; i < LoaderInfo.NumInternalChannels; i++, ptr++ )
		*ptr = NOT_PROCESS;
	Fragment[0].StartAddress = LoaderInfo.MemStart;
	Fragment[0].Len = LoaderInfo.MinInternalChannel - LoaderInfo.MemStart;
	Fragment[0].Used = 0;
	for ( i = 1; i < MAX_TASKS; i++ ){
		Fragment[i].StartAddress = NOT_PROCESS;
		Fragment[i].Len = 0;
		Fragment[i].Used = 0;
		Fragment[i].TaskId = -1;
	}
	while (1) {
		_VChanIn( &TaskInfo, ParentInChan, sizeof(TaskInfo));
		if ( TaskInfo.Action == EXIT_TASK ){
			for ( i = 0; i < MAX_TASKS; i++ )
				if ( Fragment[i].TaskId == TaskInfo.TaskId )
					break;
			if ( !Fragment[i].Used ) /* Should never occur */
				continue;
			Fragment[i].Used = 0;
			Fragment[i].TaskId = -1;
			if (( i < MAX_TASKS-1 )&& (!Fragment[i+1].Used)){
				Fragment[i].Len += Fragment[i+1].Len;
				Fragment[i+1].Len = 0;
			}
			if (( i > 0 )&& (!Fragment[i-1].Used)){
				Fragment[i-1].Len += Fragment[i].Len;
				Fragment[i].Len = 0;
			}
			continue;
		}
		else if ( TaskInfo.Action != START_TASK )
			continue;
		else {
			TaskId = -1;
			_VChanIn( &MemRequired, ParentInChan, sizeof(MemRequired));
			MemRequired = ( MemRequired / 1024 + 1)*1024;
			MemRequired += 0x1000; /* TEMP FIXME */
			for ( i = 0; i < MAX_TASKS; i++ ){
				if (( !Fragment[i].Used )&&( Fragment[i].Len >= MemRequired )){
					if (( i < MAX_TASKS )&&( Fragment[i+1].Len == 0 )){
						Fragment[i+1].Len = Fragment[i].Len - MemRequired;
						Fragment[i+1].StartAddress = Fragment[i].StartAddress + MemRequired;
						Fragment[i].Len = MemRequired;
					}
					TaskId = i;
					Fragment[i].Used = 1;
					Fragment[i].TaskId = TaskInfo.TaskId;
					break;
				}
			}
		}
		if ( TaskId < 0 ){
			ptr = 0x80000000;
			_ChanOut( &ptr, ParentOutChan, sizeof(long));
			continue;
		}
		_ChanOut( &(Fragment[TaskId].StartAddress), ParentOutChan, sizeof(long));
		_VChanIn( &CoffParameters, ParentInChan, sizeof(CoffParameters) );
		for ( i = 0; i < ((CoffParameters.BssSize > 0)?CoffParameters.NumSections-1:CoffParameters.NumSections); i++ ){
			_VChanIn( &SectionInfo, ParentInChan, sizeof(SectionInfo) );
			len = SectionInfo.SectionSize;
			addr =  SectionInfo.SectionStart;
			while ( len > 0 ){
				_VChanIn( addr, ParentInChan, len > BOOT_BUFFER_SIZE?BOOT_BUFFER_SIZE:len );
				addr += BOOT_BUFFER_SIZE;
				len -= BOOT_BUFFER_SIZE;
			}
		}
		Params = Fragment[TaskId].StartAddress + Fragment[TaskId].Len - MAX_PARAMETER_LEN;
		if ( CoffParameters.DebugFlag ) {
			_VChanIn(&len, ParentInChan, sizeof(int) );
			DebugCommand = (char *)Params;
			_VChanIn(Params, ParentInChan, len ); Params = (char *)Params + len;
		}
		_VChanIn(Params, ParentInChan, MAX_PARAMETER_LEN );
		MDL[SYS_mdl_Magic] = (void *)SYS_MAGIC_MDL;
		MDL[SYS_mdl_TextLo] = CoffParameters.TextStart;
		MDL[SYS_mdl_TextHi] = (void*)((char*) MDL[SYS_mdl_TextLo] + CoffParameters.TextSize );
		MDL[SYS_mdl_DataLo] = CoffParameters.DataStart;
		MDL[SYS_mdl_DataHi] = (void*)((char*) MDL[SYS_mdl_DataLo] + CoffParameters.DataSize + CoffParameters.BssSize );
		MDL[OS_mdl_HeapLo] = MDL[SYS_mdl_DataHi] > MDL[SYS_mdl_TextHi]?(void*)((char*) MDL[SYS_mdl_DataHi]):MDL[SYS_mdl_TextHi];
		MDL[OS_mdl_HeapHi] = (void*)((char*) MDL[OS_mdl_HeapLo] + CoffParameters.HeapSize );

		NewStack = (( int *)(MDL[OS_mdl_HeapHi]))+(CoffParameters.StackSize/4 - SECONDARY_WORKSPACE_SIZE - 10);


		/* init the stack fragment info block: */
		stackFragmentInfo[TaskId][1] = SYS_NULL_ptr;	/* Prev */
		stackFragmentInfo[TaskId][2] = SYS_NULL_ptr;	/* Next */

		stackFragmentInfo[TaskId][0] = MDL[SYS_mdl_StackLo] = (void*) (MDL[OS_mdl_HeapHi]);
		MDL[SYS_mdl_StackOvfl] = (void*) (((int *)MDL[OS_mdl_HeapHi])+80);

		MDL[SYS_mdl_StackHi] = (void*)	( &stackFragmentInfo[TaskId][3] );
		MDL[SYS_mdl_StackClass] = 0;
		MDL[OS_link_to_server] = ParentOutChan;

		FILL_PROCESS_INFO( (&ProcessInfo), 
			CoffParameters.DataStart, 
			CoffParameters.DataSize + CoffParameters.BssSize, 
			CoffParameters.TextStart, 
			CoffParameters.TextSize,
			MDL[SYS_mdl_StackHi], CoffParameters.StackSize, 
			MDL[OS_mdl_HeapLo], CoffParameters.HeapSize, 
			CoffParameters.entry, TaskId, *Params );
		ParameterBlock[0] = 7 * sizeof(long);
		ParameterBlock[1] = (unsigned long)&ProcessInfo;
		ParameterBlock[2] = CoffParameters.DebugFlag;
		ParameterBlock[3] = (unsigned long )DebugCommand;
		ParameterBlock[4] = (int)ParentInChan + 8  + 8 * TaskId;
		ParameterBlock[5] = LoaderInfo.NodeId;
		ParameterBlock[6] = LoaderInfo.TotalMem;
		ParameterBlock[7] = (int)Params;

		ptr = (int *)CoffParameters.BssStart;
		for ( i = 0; i < CoffParameters.BssSize/sizeof(int); i++ )
			*ptr++ = 0;
		for ( i = -5; i < SECONDARY_WORKSPACE_SIZE; i++ )
			NewStack[i] = OldStack[i];
		NewStack -= 4;
		NewStack[0] = 0;
		NewStack[1] = (unsigned long )MDL - (unsigned long)OldStack + (unsigned long)NewStack + 16;
		NewStack[2] = (unsigned long)&(ParameterBlock[1]) - (unsigned long)OldStack + (unsigned long)NewStack + 16;
		NewStack[3] = 0;
		StartUserTask( NewStack, CoffParameters.entry );
	}
	__asm( "stopp" );
}

